/****************************************************************************
 *
 * $Id:
 *
 * Copyright (C) 1998-2006 Lasmea. All rights reserved.
 *
 * This software was developed at:
 * Institut Pascal- UMR 6602
 * 63 175 Aubière Cedex
 * http://ip.univ-bpclermont.fr
 *
 * This file may be distributed under the terms of the Q Public License
 * as defined by Trolltech AS of Norway and appearing in the file
 * LICENSE included in the packaging of this file.
 *
 * Licensees holding valid IPDS Professional Edition licenses may
 * use this file in accordance with the IPDS Commercial License
 * Agreement provided with the Software.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Contact . if any conditions of this licensing are
 * not clear to you.
 *
 *
 * Author: Jonathan Courbon - Hemanth Korrapati -
 *
 *****************************************************************************/

#include "config.h"

#include "homogeneoustransformation.h"
#include "lasertool.h"
#include "transformationtool.h"


#if QTDETECTEDVERSION == 4
  #include <QFile>
  #include <QStringList>
  #include <QTextStream>
#else
  #include <QtCore/QFile>
  #include <QtCore/QStringList>
  #include <QtCore/QTextStream>
#endif



#include <algorithm>
#include <cassert>
#include <boost/assign/std/vector.hpp>
#include <boost/assign/std/vector.hpp>
#include <boost/algorithm/string.hpp>
#include <boost/algorithm/string/split.hpp>
#include <boost/algorithm/string/predicate.hpp>
#include <boost/algorithm/string/classification.hpp>
#include <boost/bind.hpp>

#include <fstream>
#include <cstdlib>
#include <math.h>
#include <sstream>
#include <iomanip>
#include <map>

#include <math.h>

#include "config.h"
#include "containers.h"

using namespace std;

void usage()
{
    std::cout << "=================================================================" << std::endl;
    std::cout << "  Generation of a 3D point cloud file from a trajectory and scans" << std::endl;
    std::cout << "=================================================================" << std::endl;
    std::cout << "Not any interpolation is done" << endl;
    std::cout << "Usage: ./PointCloudGeneration trajectoryFile pathtosequencefolder lidarsensorname [startingtimetraj] [output] [maxnumbofpoints]" << std::endl;
    std::cout << "  -  trajectoryFile: file containing the trajectory of the vehicle expressed in the world frame" <<endl;
    std::cout << "     Format: " << endl;
    std::cout << "       + time time x y theta (as in DeadReckonedPose) OR time x y z roll pitch yaw " << std::endl;
    std::cout << "    right-handed 'RPY'' Euler angle" << std::endl;
    std::cout << "     This is the reference for synchronisation !! " << endl;
    std::cout << "  -  lidarsensorname: hz-laser of incl-laser " << std::endl;
    std::cout << "  -  startingtimetraj: starting time of the trajectory, 0 by default " << std::endl;
    std::cout << "  -  output: name of the output file, pointcloud.xyz by default " << std::endl;
    std::cout << "  -  maxnumbofpoints: maximum number of points in the point cloud, -1 by default (i.e. no limit) " << std::endl;
}

/** \file pointCloudGeneration.cpp
 * Generation of a 3D point cloud.
 * A more elaborated file description.
 */

int main(int argc, char** argv)
{
    if(argc<4)
    {
        usage();
        return 0;
    }
    else
    {
        string sensorname=argv[3];
        if(strcmp(sensorname.c_str(),"hz-laser")!=0 && strcmp(sensorname.c_str(),"incl-laser")!=0)
        {
            usage();
            return 0;
        }
        else
        {
            string trajFile=argv[1];
            string pathtosequencefolder=argv[2];

            LaserTool laser(3.,50.);
            float zmin=1.;
            float zmax=20.;

            string output;
            long long int startingtimetraj;
            int maxnumbofpoints;

            if(argc>=5)
                startingtimetraj=atoll(argv[4]);
            else
                startingtimetraj=0llu;

            if(argc>=6)
                output=argv[5];
            else
                output="pointcloud.xyz";

            if(argc>=7)
                maxnumbofpoints=atoi(argv[6]);
            else
                maxnumbofpoints=-1;

            cout << "=========================" << endl;
            cout << "The file containing the 3D point cloud: " << output << " will be generated using the lidar "<< sensorname << " and the trajectory of the vehicle given in the file: "<< trajFile << " with a time offset of "<< startingtimetraj << " us with a maximum number of points of "<< maxnumbofpoints << endl;
            cout << "=========================" << endl;




            map<string,HomogeneousTransformation> map_t_vehicleToLidars;
            map_t_vehicleToLidars.insert(pair<string,HomogeneousTransformation>("hz-laser",HomogeneousTransformation(1.5,0.,0.22,-M_PI/2.,0.,0)));
            map_t_vehicleToLidars.insert(pair<string,HomogeneousTransformation>("incl-laser",HomogeneousTransformation(1.505, 0.1, 1.96, -M_PI/2., 0.*M_PI/180., -20.*M_PI/180.)));

            // read the Ctable
            ConfigTable ctable;
            std::stringstream msg;
            msg << CONFIGFOLDERPATH << "/config_table";
            ctable.readTable(msg.str());
            // read the lidar file
            string timestamps_file_name=ctable.table[sensorname].folderName+ctable.table[sensorname].suffix;
            string sensor_data_folder=pathtosequencefolder+"/"+ctable.table[sensorname].folderName+"/";
            std::vector<Unit> unitsLidar;
            getReadings(ctable.table[sensorname].sensorName, sensor_data_folder, timestamps_file_name, ctable.table[sensorname].tfm, unitsLidar, TimestampInterval(),1);

            cout << "There are "<< unitsLidar.size() << " elements in the lidar list"<< endl;
            if(unitsLidar.size()!=0)
                cout << "   First element starts at: " << unitsLidar[0].timestamp << endl;

            // generate the units for the trajectory
            QFile file(trajFile.c_str());
            if (!file.open(QIODevice::ReadOnly | QIODevice::Text))
            {
                cout << "The file " << trajFile << " is not valid" << endl;
                return 0;
            }
            // read file line by line
            std::vector<Unit> unitsTraj;
            Unit unite;
            QStringList list;
            vector<string> reading(6);

            setlocale (LC_ALL,"C");
            int ind=-1;
            QTextStream in(&file);
            while (!in.atEnd()) {
                QString line = in.readLine();
                list=line.split(" ");
                if(list.size()==4)
                {
                    reading[0]=list[1].toStdString();
                    reading[1]=list[2].toStdString();
                    reading[2]=0.;
                    reading[3]=list[3].toStdString();
                    reading[4]=0.;
                    reading[5]=0.;
                }
                else if(list.size()==7)
                {
                    for(int i=0;i<6;i++)
                    {
                        reading[i]=list[i+1].toStdString();
                    }
                }
                else
                {
                    cout << "The content of the file " << trajFile << " is not valid, must be time x y yaw or time x y z r p y" << endl;
                    return 0;
                }
                ind++;
                unite.timestamp=atoll(list[0].toStdString().c_str())+startingtimetraj;
                unite.index=ind;
                unite.reading=reading;
                unitsTraj.push_back(unite);
            }

            cout << "There are "<< unitsTraj.size() << " elements in the trajectory file"<< endl;
            if(unitsTraj.size()!=0)
                cout << "    First element starts at: " << unitsTraj[0].timestamp << endl;

            if(unitsLidar.size()!=0 && unitsTraj.size()!=0)
            {
                HomogeneousTransformation t_worldToVehicle,t_worldToLidar;
                std::vector<HomogeneousVector> vecImpacts;
                string filepath;


                // synchronise
                vector<int> synchroMap;
                synchronizePair(unitsTraj, unitsLidar, synchroMap);

                ofstream sortie;
                sortie.open(output.c_str());


                int nbmatch=0;
                int nbpoints=0;
                int nbpointsfil=0;
                bool addpoint=true;

                unsigned int i=0;
                while(i<synchroMap.size() && addpoint)
                {
                    if(synchroMap[i]!=-1) // synchro OK
                    {
                        nbmatch++;
                        // get the frame t_worldToVehicle
                        t_worldToVehicle=HomogeneousTransformation(atof(unitsTraj[i].reading[0].c_str()),atof(unitsTraj[i].reading[1].c_str()),atof(unitsTraj[i].reading[2].c_str()),atof(unitsTraj[i].reading[3].c_str()),atof(unitsTraj[i].reading[4].c_str()),atof(unitsTraj[i].reading[5].c_str()));
                        t_worldToLidar=t_worldToVehicle*map_t_vehicleToLidars[sensorname];
                        // get the impact file
                        ctable.getSensorFilePathFromIndex(sensorname,filepath,synchroMap[i]);
                        filepath=pathtosequencefolder+"/"+filepath;
                        //cout << "filepath= "<< filepath << endl;
                        TransformationTool::expressImpactInFrame(laser,filepath,t_worldToLidar,vecImpacts);

                        sortie << t_worldToLidar.Gett1() << " " << t_worldToLidar.Gett2() << " "<< t_worldToLidar.Gett3() << endl;


                        for(unsigned int j=0;j<vecImpacts.size();j++)
                        {
                            if(addpoint)
                            {
                                if(vecImpacts[j].GetZ()>=zmin && vecImpacts[j].GetZ()<=zmax)
                                {
                                    sortie << vecImpacts[j].GetX()<< " " << vecImpacts[j].GetY()<< " " << vecImpacts[j].GetZ() << endl;
                                    nbpointsfil++;
                                }

                                nbpoints++;

                                if(nbpoints%100000==0)
                                    cout << nbpoints << " points added (without filtering)" << endl;

                                if(maxnumbofpoints!=-1)
                                {
                                    if(nbpoints>=maxnumbofpoints)
                                    {
                                        cout << " maximum number of points reached ..." << endl;
                                        addpoint=false;
                                    }
                                }

                            }
                        }

                    }
                    i++;
                }

                cout << nbmatch << " readings synchronised over "<< synchroMap.size()<< " and "<< nbpoints << " 3D points obtained / " << nbpointsfil << " points after altitude filtering ( "<< zmin << " < z < "<< zmax << " ; the number of points in the point cloud)" << endl;

            }

        }

    }
}

